/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for Petit FatFs (C)ChaN, 2009      */
/*-----------------------------------------------------------------------*/
//////////////////////////////////////////////
// HEADER FILES
//////////////////////////////////////////////
#include "spi.h"
#include "diskio.h"
//////////////////////////////////////////////
// Definitions for MMC/SDC command 
//////////////////////////////////////////////
#define CMD0	(0x40+0)	/* GO_IDLE_STATE */
#define CMD1	(0x40+1)	/* SEND_OP_COND (MMC) */
#define	ACMD41	(0xC0+41)	/* SEND_OP_COND (SDC) */
#define CMD8	(0x40+8)	/* SEND_IF_COND */
#define CMD9	(0x40+9)	/* SEND_CSD */
#define CMD10	(0x40+10)	/* SEND_CID */
#define CMD12	(0x40+12)	/* STOP_TRANSMISSION */
#define ACMD13	(0xC0+13)	/* SD_STATUS (SDC) */
#define CMD16	(0x40+16)	/* SET_BLOCKLEN */
#define CMD17	(0x40+17)	/* READ_SINGLE_BLOCK */
#define CMD18	(0x40+18)	/* READ_MULTIPLE_BLOCK */
#define CMD23	(0x40+23)	/* SET_BLOCK_COUNT (MMC) */
#define	ACMD23	(0xC0+23)	/* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24	(0x40+24)	/* WRITE_BLOCK */
#define CMD25	(0x40+25)	/* WRITE_MULTIPLE_BLOCK */
#define CMD55	(0x40+55)	/* APP_CMD */
#define CMD58	(0x40+58)	/* READ_OCR */
//////////////////////////////////////////////
// QUICK FUNCTIONS
//////////////////////////////////////////////
#define CLOCKSPI()  writeSPI(0xFF)
#define SELECT()	MMC_CS=0
#define DESELECT()	MMC_CS=1;CLOCKSPI()

BYTE CardType;
//////////////////////////////////////////////
// SEND COMMAND
//////////////////////////////////////////////
BYTE send_cmd (BYTE cmd, DWORD arg){
	BYTE n, res;
    BYTE i;

    // If its a ACMD the command sequense of CMD55 then (ACMD & 0x7F) 
	if (cmd & 0x80) {	
		cmd &= 0x7F;
		res = send_cmd(CMD55, 0);
		if (res > 1) return res;
	}

	/* Select the card and wait for ready */
	DESELECT();
	SELECT();
	if (sd_wait_ready() != 0xFF) return 0xFF;

	/* Send command packet */
	writeSPI(cmd);						/* Start + Command index */
	writeSPI((BYTE)(arg >> 24));		/* Argument[31..24] */
	writeSPI((BYTE)(arg >> 16));		/* Argument[23..16] */
	writeSPI((BYTE)(arg >> 8));			/* Argument[15..8] */
	writeSPI((BYTE)arg);				/* Argument[7..0] */
	n = 0x01;							/* Dummy CRC + Stop */
	if (cmd == CMD0) n = 0x95;			/* Valid CRC for CMD0(0) */
	if (cmd == CMD8) n = 0x87;			/* Valid CRC for CMD8(0x1AA) */
	writeSPI(n);

	/* Receive command response */
	if (cmd == CMD12) rcv_spi();	/* Skip a stuff byte when stop reading */

    for( i=0; i<10; i++) 
    {
        res = rcv_spi();      
        if ( res != 0xFF) 
            break;
    }


	return res;			/* Return with the response value */
}
//////////////////////////////////////////////
// WAIT UNTIL SD CARD IS READY
//////////////////////////////////////////////
BYTE sd_wait_ready (void){
	BYTE res;
    WORD time,i;

	time = 10000 ;	        //EDIT THIS: For 40Mhz thats 10MIPS for 1/2 Second thats 5,000,000 Cycles
                            //One must setup a timer for 500ms but since i hate times i just loop
                            //about 10,000 times

    for (i=0; i<time; i++) 
    {
        res = rcv_spi();
        if (res == 0xFF) 
            break; 
    } 

	return res;
}
//////////////////////////////////////////////
// Get DISK Status
//////////////////////////////////////////////
DSTATUS disk_status (void)
{
	DSTATUS stat;

	// Put your code here

	return stat;
}
//////////////////////////////////////////////
// Read Partial Sector 
//////////////////////////////////////////////
DRESULT disk_readp (
	BYTE *dest,		/* Pointer to the destination object to put data */
	DWORD lba,		/* Start sector number (LBA) */
	WORD ofs,		/* Byte offset in the sector (0..511) */
	WORD cnt		/* Byte count (1..512), b15:destination flag */
)
{

	DRESULT res;
	BYTE *p, dev, rc;
	WORD cf, tmr;
	BYTE *f;


	dev = (cnt & 0x8000) ? 1 : 0;	// Destination type 
	cnt &= 0x7FFF;
	if (!cnt || ofs + cnt > 512) return RES_PARERR;

	if (!(CardType & CT_BLOCK)) lba *= 512;		// Convert to byte address if needed 

	res = RES_ERROR;
	if (send_cmd(CMD17, lba) == 0) {		// READ_SINGLE_BLOCK 

		tmr = 1000;
		do {							// Wait for data packet in timeout of 200ms 
			rc = rcv_spi();
		} while (rc == 0xFF && --tmr);

		if (rc == 0xFE) {

			cf = 512 + 2 - ofs - cnt;

			while (ofs--) rcv_spi();	// Skip top of sector 

			if (dev) {					/* Receive middle of the sector */
				f = dest;
				do
					res = rcv_spi();
				while (--cnt && res);
				cf += cnt;
			} else {
				p = dest;
				do
					*p++ = rcv_spi();
				while (--cnt);
			}

			do							// Skip bottom of sector and CRC 
				rcv_spi();
			while (--cf);

			res = cnt ? RES_STRERR : RES_OK;
		}
	}

	release_spi();

	return cnt ? RES_ERROR : RES_OK;

}
//////////////////////////////////////////////
// Initialize Disk Drive   
//////////////////////////////////////////////
DSTATUS disk_initialize (void)
{
    BYTE SDHC = 0;
    BYTE ocr[4] = {0,0,0,0};
    BYTE ty,cmd;

	WORD tmr;
    WORD r,i,x;
    
    tmr = 10000;
    r = 0;

    init_spi();

    // 1. with the card NOT selected
    DESELECT();

    // 2. send 80 clock cycles start up
    for ( i=0; i<10; i++)
        writeSPI(0xFF);

    // 3. now select the card
    SELECT();

    // 4. send a single RESET command
    r = send_cmd(CMD0, 0); 
    //DESELECT();
    if ( r != 1)                // must return Idle
        return 1;   // comand rejected

    // 5. send repeatedly INIT until Idle terminates
    if (send_cmd(CMD8, 0x1AA) == 1) {	//* SDC ver 2.00 
        SDHC = 1;
        i = 0;
        for (x = 0; x < 4; x++) ocr[x] = rcv_spi();
		if (ocr[2] == 0x01 && ocr[3] == 0xAA) {	                    //* The card can work at vdd range of 2.7-3.6V 
            for (i=0; i<tmr; i++){
                r = send_cmd(ACMD41, 1UL << 30); 
                DESELECT();
                if (!r) 
                    break; 
            }
		    if (send_cmd(CMD58, 0) == 0) {	                        // Check CCS bit 
			    for (x = 0; x < 4; x++) ocr[x] = rcv_spi();
			    ty = (ocr[0] & 0x40) ? CT_SD2|CT_BLOCK : CT_SD2;	// SDv2 
		    }
	    }
    }else{

        if (send_cmd(ACMD41, 0) <= 1) {
        	ty = CT_SD1;
            cmd = ACMD41;	//* SDv1 
		} else {
			ty = CT_MMC; 
            cmd = CMD1;	//* MMCv3 
		}

        for (i=0; i<tmr; i++){
            r = send_cmd(cmd, 0); DESELECT();
            if ( !r) 
                break; 
        } 
        if ( i == tmr)   
            return 2;  // init timed out 


    }
	return r;
}